package aerogen;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Random;

import org.bukkit.Material;
import org.bukkit.TreeType;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.util.noise.SimplexOctaveGenerator;

public class Island extends Schematic {
	
	public static Island createIsland(AeroGenerator gen, World world, Random random, int chunkX, int chunkZ) {
		
		Biome mcBiome = world.getBiome(chunkX*16+8, chunkZ*16+8);
		BiomeDescription biome = BiomeDescription.getDescription(mcBiome);
		
		if(!biome.doCreateIsland(random) && !(chunkX == 0 && chunkZ == 0)) return null;
		Int3D size = null;
		Int3D position = null;
		if(chunkX == 0 && chunkZ == 0) {
			biome = BiomeDescription.getDescription(Biome.FOREST);
			size = new Int3D(128, 64, 128);
			position = new Int3D(-56, 48, -56);
			gen.spawn = new Int3D(0, (int) (position.y+size.y-biome.roughness), 0);
		} else {
			size = biome.getDimensions(random);
			position = biome.getPosition(random, chunkX, chunkZ, size);
		}
		
		return new Island(biome, random, position, size);
		
	}
	
	private static final byte grass = (byte) Material.GRASS.getId();
	private static final byte dirt = (byte) Material.DIRT.getId();
	private static final byte stone = (byte) Material.STONE.getId();
	private static final byte air = (byte) Material.AIR.getId();
	private static final byte snow = (byte) Material.SNOW.getId();
	private static final byte sand = (byte) Material.SAND.getId();
	private static final byte gravel = (byte) Material.GRAVEL.getId();
	private static final byte water = (byte) Material.WATER.getId();
	private static final byte lava = (byte) Material.LAVA.getId();
	private static final byte reed = (byte) Material.SUGAR_CANE_BLOCK.getId();
	private static final byte cactus = (byte) Material.CACTUS.getId();
	private static final byte flowerRed = (byte) Material.RED_ROSE.getId();
	private static final byte flowerYellow = (byte) Material.YELLOW_FLOWER.getId();
	private static final byte mushroomBrown = (byte) Material.BROWN_MUSHROOM.getId();
	private static final byte mushroomRed = (byte) Material.RED_MUSHROOM.getId();
	private static final byte pumpkin = (byte) Material.PUMPKIN.getId();
	private static final byte watermelon = (byte) Material.MELON_BLOCK.getId();
		
	protected int centerX;
	protected int centerY;
	protected int centerZ;
	protected Random random;
	protected boolean[][] isLand;
	protected int[][] distanceFromEdge;
	protected int[][] heightMap;
	protected int[][] bottomMap;
	protected int[][] cliffMap;
	protected double[][] heightCoefficients;
	protected BiomeDescription biome;
	protected int islandMass = 0;
	
	public Island(BiomeDescription biome, Random random, Int3D position, Int3D size) {
		super(position.x, position.y, position.z, size.x, size.y, size.z);
		
		this.biome = biome;
		
		this.random = random;
		
		this.centerX = this.length / 2;
		this.centerZ = this.bredth / 2;
		this.centerY = this.height - biome.getAirSpace();
		
		generateIslandForm();			//Generates topographical shape
		calculateHeightCoefficients();	//Calculates lots of useful data
		generateIslandTerrain();		//Main terrain generation
		erode();						//Roughens up the edges of things and makes them look nice
		addHollows();					//Caves and water/lava ponds
		populatePlants();				//Trees and shrubs + support falling materials
		addExtras();					//Lava falls, water falls, and ores
		
	}

	/**
	 * Generates a blob-like surface shape for the island using
	 * a technique that produces random blobs followed by an
	 * after-pass to clean up small holes.
	 */
	private void generateIslandForm() {
				
		this.isLand = new boolean[length][bredth];
		
		ArrayList<Point> points = new ArrayList<Point>();
		points.add(new Point(centerX-1, centerZ));
		points.add(new Point(centerX+1, centerZ));
		points.add(new Point(centerX, centerZ-1));
		points.add(new Point(centerX, centerZ+1));
		isLand[centerX][centerZ] = true;
		isLand[centerX-1][centerZ] = true;
		isLand[centerX+1][centerZ] = true;
		isLand[centerX][centerZ-1] = true;
		isLand[centerX][centerZ+1] = true;
		
		int num = (int) (length*bredth*0.5);
		
		double xOdds = (double)length / (double)(length+bredth);
		double zOdds = 1.0-xOdds;
		
		for(int i = 0; i < num && points.size() > 0; i++) {
			
			Point p = points.get(random.nextInt(points.size()));
			
			int x = (int) p.getX();
			int z = (int) p.getY();
			
			if(x > 0 && !isLand[x - 1][z] && random.nextDouble() < xOdds) {
				Point q = new Point(x - 1, z);
				if(!points.contains(q)) {
					if(x > 1) points.add(q);
					isLand[x - 1][z] = true;
				}
			}
			
			if(x < length - 1 && !isLand[x + 1][z] && random.nextDouble() < xOdds) {
				Point q = new Point(x + 1, z);
				if(!points.contains(q)) {
					if(x < length - 2) points.add(q);
					isLand[x + 1][z] = true;
				}
			}
			
			if(z > 0 && !isLand[x][z - 1] && random.nextDouble() < zOdds) {
				Point q = new Point(x, z - 1);
				if(!points.contains(q)) {
					if(z > 1) points.add(q);
					isLand[x][z - 1] = true;
				}
			}
			
			if(z < bredth - 1 && !isLand[x][z + 1] && random.nextDouble() < zOdds) {
				Point q = new Point(x, z + 1);
				if(!points.contains(q)) {
					if(z < bredth - 2) points.add(q);
					isLand[x][z + 1] = true;
				}
			}
			
			if(isLand[x-1][z] && isLand[x+1][z] && isLand[x][z-1] && isLand[x][z+1]) {
				points.remove(p);
			}
			
		}

		points.clear();
		
		for(int x = 1; x < length - 1; x++) {
			for(int z = 1; z < bredth - 1; z++) {
				if(!isLand[x][z]) {
					int n = 0;
					if(isLand[x - 1][z]) n++;
					if(isLand[x + 1][z]) n++;
					if(isLand[x][z - 1]) n++;
					if(isLand[x][z + 1]) n++;
					if(n > 1) points.add(new Point(x, z));
				}
			}
		}
		
		for(Point p : points) {
			int x = (int) p.getX();
			int z = (int) p.getY();
			if(x >= 0 && x < length && z >= 0 && z < bredth) {
				isLand[x][z] = true;
			}
		}
	}

	/**
	 * Calculates each tile in the islands Manhattan distance from
	 * the nearest edge of the island, then normalizes and curves it
	 * to create a nice coefficient to multiply heights by to make an
	 * island with smooth edges.
	 */
	private void calculateHeightCoefficients() {
		
		this.heightCoefficients = new double[length][bredth];
		this.distanceFromEdge = new int[length][bredth];
		
		for(int x = 0; x < length; x++) {
			for(int z = 0; z < bredth; z++) {
				if(!isLand[x][z]) {
					heightCoefficients[x][z] = 0;
				} else {
					heightCoefficients[x][z] = 1;
				}
			}
		}
		
		int edge = (length + bredth) / 4;
		double max = 0;
		for(int i = 0; i < edge; i++) {
			for(int x = 1; x < length - 1; x++) {
				for(int z = 1; z < bredth - 1; z++) {
					if(heightCoefficients[x][z] > 0) {
						
						double lowNeighbor = 100;
						
						if(heightCoefficients[x - 1][z] < lowNeighbor) lowNeighbor = heightCoefficients[x - 1][z];
						if(heightCoefficients[x + 1][z] < lowNeighbor) lowNeighbor = heightCoefficients[x + 1][z];
						if(heightCoefficients[x][z - 1] < lowNeighbor) lowNeighbor = heightCoefficients[x][z - 1];
						if(heightCoefficients[x][z + 1] < lowNeighbor) lowNeighbor = heightCoefficients[x][z + 1];
						
						heightCoefficients[x][z] = lowNeighbor + 1;
						if(lowNeighbor + 1 > max) max = lowNeighbor + 1;
						
					}
				}
			}
		}
		
		for(int x = 0; x < length; x++) {
			for(int z = 0; z < bredth; z++) {
				distanceFromEdge[x][z] = (int) heightCoefficients[x][z];
				heightCoefficients[x][z] = heightCoefficients[x][z] / max;
				heightCoefficients[x][z] = 1.0 - Math.pow(1.0 - heightCoefficients[x][z], 2.5); //Power adjustment makes it look nicer
			}
		}
		
	}

	private void generateIslandTerrain() {
		boolean lake = false;
		if(biome.doLake(random)) lake = true;
		int cliffHeight = 0;
		if(biome.cliffs) {
			cliffHeight = biome.rollCliffHeight(random);
		}
		heightMap = new int[length][bredth];
		bottomMap = new int[length][bredth];
		cliffMap = new int[length][bredth];
		
		SimplexOctaveGenerator gen = new SimplexOctaveGenerator(random, 8);
		gen.setScale(1.0 / 64.0);
		
		for(int x = 0; x < length; x++) {
			for(int z = 0; z < bredth; z++) {
				if(isLand[x][z]) {
					double distance = heightCoefficients[x][z];
					double noiseVal = gen.noise(x, z, biome.frequency, 0.25);
					double maxYd = noiseVal * Math.pow(distance, 0.2);
					if(lake) maxYd *= distance;
					maxYd += 0.5;
					if(maxYd < 0.0) maxYd = 0.0;
					int maxY = (int) (maxYd * biome.roughness);
					maxY += centerY;

					if(biome.snow && maxY > height - 2) maxY = height - 2;
					if(maxY > height - 1) {
						maxY = height - 1;
					}
					if(maxY < 0) maxY = 0;
					heightMap[x][z] = maxY;
					
					int cliff = 0;
					if(biome.cliffs) {
						double cliffnoise = gen.noise(x, z, 0.4, 0.25);
						cliffnoise -= 0.5*biome.lakeChance*biome.lakeChance*biome.lakeChance;
						if(cliffnoise > 0 && distance > 0.9) cliff = cliffHeight;
						maxY += cliff;
					}
					cliffMap[x][z] = cliff;
					
					if(biome.snow && maxY > height - 2) maxY = height - 2;
					if(maxY > height - 1) {
						maxY = height - 1;
					}
					if(maxY < 0) maxY = 0;
					
					int minY = (int) (centerY * (1.0-(0.8+gen.noise(x, z, 3.0, 0.25)*0.4)*distance));
					if(minY > centerY-4) minY = (int) (centerY-4+gen.noise(x, z, 3.0, 0.25)*2*distance);
					if(minY < 0) minY = 0;
					bottomMap[x][z] = minY;					
					
					int range = maxY - minY;
					int soilDepth = (int) (range * (1.0-biome.islandSoilDepth));
					double sedimentNoise = gen.noise(x, z, 0.8, 0.25)+0.5;
					for(int y = minY; y < maxY; y++) {
						islandMass++;
						if((y - minY) < soilDepth || cliff > 0) {
							setBlock(x, y, z, biome.islandRockMaterial);
						} else {
							setBlock(x, y, z, biome.islandSoilMaterial);
						}
					}
					
					if(biome.snow && !lake) setBlock(x, maxY + 1, z, biome.islandSnowMaterial);

					if(lake && maxY < centerY+biome.roughness/2) {
						if(sedimentNoise < 0.33) {
							setBlock(x, maxY, z, biome.islandLakeBedDirt);
						} else {
							setBlock(x, maxY, z, biome.islandLakeBedMaterial);
						}
						for(int y = maxY + 1; y < centerY+cliff; y++) {
							setBlock(x, y, z, biome.islandFluidMaterial);
						}
					} else {
						setBlock(x, maxY, z, biome.islandSurfaceMaterial);
					}
				}
			}
		}
	}
	
	private void erode() {
		for(int i = 1; i < length-1; i++) {
			for(int k = 1; k < bredth-1; k++) {
				if(distanceFromEdge[i][k] == 1) {
					int jm = random.nextInt(heightMap[i][k] - bottomMap[i][k] + 1) + bottomMap[i][k];
					for(int j = 0; j < jm; j++) {
						byte block = getBlock(i, j, k);
						if(block == biome.islandRockMaterial || block == biome.islandSoilMaterial) {
							setBlock(i, j, k, air);
						}
					}
				}
				if(cliffMap[i][k] > 0) {
					if(cliffMap[i-1][k] == 0 || cliffMap[i+1][k] == 0 || cliffMap[i][k-1] == 0 || cliffMap[i][k+1] == 0) {
						int jm = random.nextInt(cliffMap[i][k]*3/4) + heightMap[i][k];
						for(int j = heightMap[i][k]+1; j < jm; j++) {
							byte block = getBlock(i, j, k);
							if(block == biome.islandRockMaterial) {
								setBlock(i, j, k, air);
							}
						}
					}
				}
			}
		}
	}
	
	private void populatePlants() {
		for(int i = 1; i < length-1; i++) {
			for(int k = 1; k < bredth-1; k++) {
				int j = heightMap[i][k]+cliffMap[i][k];
				if(j < height-1 && getBlock(i, j+1, k) == air || getBlock(i, j+1, k) == snow) {
					if(getBlock(i, j, k) == grass) {
						if(biome.rollTree(random)) {
							generateTree(i, j+1, k, biome.rollTreeType(random));
							continue;
						}
						if(getBlock(i+1, j, k) == water || getBlock(i-1, j, k) == water || getBlock(i, j, k+1) == water || getBlock(i, j, k-1) == water) {
							if(biome.rollReed(random)) {
								placeReed(i, j+1, k, random.nextInt(3)+1);
								continue;
							}
						}
						if(biome.rollFlower(random)) {
							if(biome.rollFlowerType(random)) {
								placePlantCluster(flowerRed, i, k, biome.rollFlowerClusterSize(random), 5);
								continue;
							} else {
								placePlantCluster(flowerYellow, i, k, biome.rollFlowerClusterSize(random), 5);		
								continue;						
							}
						}
						if(biome.rollMushroom(random)) {
							if(biome.rollMushroomType(random)) {
								placePlantCluster(mushroomRed, i, k, biome.rollMushroomClusterSize(random), 5);
								continue;
							} else {
								placePlantCluster(mushroomBrown, i, k, biome.rollMushroomClusterSize(random), 5);			
								continue;					
							}
						}
						if(biome.rollPumpkin(random)) {
							placePlantCluster(pumpkin, i, k, biome.rollPumpkinClusterSize(random), 5);
							continue;
						}
						if(biome.rollWatermelon(random)) {
							placePlantCluster(watermelon, i, k, biome.rollWatermelonClusterSize(random), 5);
							continue;
						}
						if(biome.rollShrub(random)) {
							byte type = biome.rollShrubType(random);
							if(type != 0) placeShrub(i, j+1, k, type);
							continue;
						}
					} else if(getBlock(i, j, k) == sand) {
						if(getBlock(i+1, j, k) == water || getBlock(i-1, j, k) == water || getBlock(i, j, k+1) == water || getBlock(i, j, k-1) == water) {
							if(biome.rollReed(random)) {
								placeReed(i, j+1, k, random.nextInt(3)+1);
								continue;
							}
						}
						if(getBlock(i+1, j+1, k) == air || getBlock(i-1, j+1, k) == air || getBlock(i, j+1, k+1) == air || getBlock(i, j+1, k-1) == air) {
							if(biome.rollCactus(random)) {
								placeCactus(i, j+1, k, random.nextInt(3)+1);
								continue;
							}
						}
						if(biome.rollShrub(random)) {
							byte type = biome.rollShrubType(random);
							if(type == 0) placeShrub(i, j+1, k, type);
							continue;
						}
					}
				}
				//Plant grass and support falling materials
				for(j = bottomMap[i][k]; j < heightMap[i][k]+cliffMap[i][k]+2 && j < height; j++) {
					if(getBlock(i, j, k) == sand || getBlock(i, j, k) == gravel) {
						if(j > 0) {
							if(getBlock(i, j-1, k) == air) setBlock(i, j-1, k, biome.islandRockMaterial);
						} else {
							setBlock(i, j, k, biome.islandRockMaterial);
						}
					}
					if(getBlock(i, j, k) == water || getBlock(i, j, k) == lava) {
						if(j > 0) {
							if(getBlock(i, j-1, k) == air) setBlock(i, j-1, k, biome.islandRockMaterial);
							if(getBlock(i-1, j, k) == air) setBlock(i-1, j, k, biome.islandSoilMaterial);
							if(getBlock(i+1, j, k) == air) setBlock(i+1, j, k, biome.islandSoilMaterial);
							if(getBlock(i, j, k-1) == air) setBlock(i, j, k-1, biome.islandSoilMaterial);
							if(getBlock(i, j, k+1) == air) setBlock(i, j, k+1, biome.islandSoilMaterial);
						} else {
							setBlock(i, j, k, biome.islandRockMaterial);
						}
					}
					if(biome.islandSurfaceMaterial == grass && j < height && getBlock(i, j, k) == dirt && getBlock(i, j+1, k) == air) {
						setBlock(i, j, k, grass);
					}
				}
			}
		}
	}
	
	private void generateTree(int x, int y, int z, TreeType type) {
		switch(type) {
			case BIG_TREE: 
				generateBigTree(x, y, z);
				return;
			case BIRCH:
				generateBirchTree(x, y, z);
				return;
			case BROWN_MUSHROOM:
				generateTallBrownMushroom(x, y, z);
				return;
			case RED_MUSHROOM:
				generateTallRedMushroom(x, y, z);
				return;
			case REDWOOD:
				generateRedwoodTree(x, y, z);
				return;
			case TALL_REDWOOD:
				generateTallRedwoodTree(x, y, z);
				return;
			case TREE:
				generateTree(x, y, z);
				return;
		}
	}
	
	private void generateTree(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.BRICK.getId());
	}
	
	private void generateBigTree(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.BEDROCK.getId());
	}
	
	private void generateBirchTree(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.DIAMOND_BLOCK.getId());
	}
	
	private void generateRedwoodTree(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.LAPIS_BLOCK.getId());
	}
	
	private void generateTallRedwoodTree(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.SPONGE.getId());
	}
	
	private void generateTallBrownMushroom(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.GOLD_BLOCK.getId());
	}
	
	private void generateTallRedMushroom(int x, int y, int z) {
		setBlock(x, y, z, (byte) Material.IRON_BLOCK.getId());
	}
	
	private void placeShrub(int x, int y, int z, byte type) {
		switch(type) {
			case 0:
				setBlock(x, y, z, (byte) Material.CHEST.getId());
				return;
			case 1:
				setBlock(x, y, z, (byte) Material.WORKBENCH.getId());
				return;
			case 2:
				setBlock(x, y, z, (byte) Material.FURNACE.getId());
				return;
		}
	}
	
	private void placeReed(int x, int y, int z, int height) {
		for(int j = 0; j < height; j++) {
			setBlockSafe(x, y+j, z, reed);
		}
	}
	
	private void placeCactus(int x, int y, int z, int height) {
		for(int j = 0; j < height; j++) {
			if(getBlockSafe(x-1, y, z) == air && getBlockSafe(x+1, y, z) == air && getBlockSafe(x, y, z-1) == air && getBlockSafe(x, y, z+1) == air) {
				setBlockSafe(x, y+j, z, cactus);
			}
		}
	}
	
	private void placePlantCluster(byte type, int x, int z, int numPlants, int radius) {
		placePlant(type, x, z);
		for(int i = 1; i < numPlants; i++) {
			placePlant(type, x+random.nextInt(radius*2)-radius, z+random.nextInt(radius*2)-radius);
		}
	}

	private void placePlant(byte type, int x, int z) {
		if(x < 0 || x >= length || z < 0 || z >= bredth) return;
		int y = heightMap[x][z]+cliffMap[x][z];
		if(getBlockSafe(x, y, z) == grass) setBlockSafe(x, y+1, z, type);
	}
	
	private void addHollows() {
		
		Random random = new Random(this.random.nextLong());
		int x = length/2, y = height/2, z = bredth/2;
		
		double angle = random.nextDouble()*Math.PI*2.0;
		CaveNode nodeL = new CaveNode(x, y, z, angle, -0.2, biome.caveDescription.startRadius);
		CaveSystem systemL = new CaveSystem(this, random, nodeL, biome.getMaximumCaveSpace(islandMass)/2);
		CaveNode nodeR = new CaveNode(x, y, z, angle-Math.PI, -0.2, biome.caveDescription.startRadius);
		CaveSystem systemR = new CaveSystem(this, random, nodeR, biome.getMaximumCaveSpace(islandMass)/2);
		
		for(int i = 0; i < length; i++) {
			for(int j = 0; j < height; j++) {
				for(int k = 0; k < bredth; k++) {
					if(systemL.contains(i, j, k)) {
						setBlock(i, j, k, caveSpace);
					} else if(systemR.contains(i, j, k)) {
						setBlock(i, j, k, caveSpace);
					}
				}
			}
		}
		
		for(int i = 0; i < length; i++) {
			for(int j = 0; j < height; j++) {
				for(int k = 0; k < bredth; k++) {
					if(getBlock(i, j, k) == biome.islandSurfaceMaterial && biome.rollWaterPond(random)) {
						addLake(i, j, k, biome.islandFluidMaterial);
					}
					if(getBlock(i, j, k) == biome.islandRockMaterial && biome.rollLavaPond(random)) {
						addLake(i, j, k, lava);
					}
				}
			}
		}
		
	}
	
	private void addLake(int x, int y, int z, byte fluid) {
		boolean[][] lake = new boolean[5][5];
		lake[2][2] = true;
		for(int j = 0; j < 4; j++) {
			for(int i = 0; i < 5; i++) {
				for(int k = 0; k < 5; k++) {
					if(i > 0 && lake[i-1][k] && random.nextDouble() < 0.5) {
						lake[i-1][k] = true;
						continue;
					}
					if(i < 4 && lake[i+1][k] && random.nextDouble() < 0.5) {
						lake[i+1][k] = true;
						continue;
					}
					if(k > 0 && lake[i][k-1] && random.nextDouble() < 0.5) {
						lake[i][k-1] = true;
						continue;
					}
					if(k < 4 && lake[i][k+1] && random.nextDouble() < 0.5) {
						lake[i][k+1] = true;
						continue;
					}
				}
			}
		}
		for(int i = 0; i < 5; i++) {
			for(int k = 0; k < 5; k++) {
				if(lake[i][k]) {
					setBlockSafe(x-2+i, y-1, z-2+k, fluid);
					setBlockSafe(x-2+i, y, z-2+k, fluid);
					setBlockSafe(x-2+i, y+1, z-2+k, air);
				}
			}
		}
	}

	private void addExtras() {
		for(int i = 1; i < length-1; i++) {
			for(int k = 1; k < bredth-1; k++) {
				for(int j = bottomMap[i][k]; j <= heightMap[i][k]+cliffMap[i][k]; j++) {
					byte block = getBlock(i, j, k);
					if(block == biome.islandRockMaterial || block == biome.islandSoilMaterial) {
						int n = 0;
						if(getBlockSafe(i, j-1, k) != caveSpace && getBlockSafe(i, j+1, k) != caveSpace) { 
							if(getBlock(i+1, j, k) == caveSpace) n++;
							if(getBlock(i-1, j, k) == caveSpace) n++;
							if(n < 2 && getBlock(i, j, k+1) == caveSpace) n++;
							if(n < 2 && getBlock(i, j, k-1) == caveSpace) n++;
							if(n == 1 && random.nextDouble() < biome.liquidFallChance) {
								if(random.nextDouble() < biome.lavaFallChance) {
									setBlock(i, j, k, lava);
								} else {
									setBlock(i, j, k, water);
								}
								continue;
							}
						}
					}
					if(block == stone) {
						double depth = (double)(j-bottomMap[i][k]) / (double)(heightMap[i][k]+cliffMap[i][k]-bottomMap[i][k]);
						int oreVein = biome.rollOreVein(random, depth);
						if(oreVein > -1) {
							int amount = biome.rollOreVeinSize(random, oreVein);
							byte material = biome.getOreMaterial(oreVein);
							populateOreVein(i, j, k, material, amount);
						}
					}
				}
			}
		}
	}

	private void populateOreVein(int i, int j, int k, byte material, int amount) {
		
		ArrayList<Int3D> nodes = new ArrayList<Int3D>();
		ArrayList<Int3D> neighbors = new ArrayList<Int3D>();
		nodes.add(new Int3D(i, j, k));
		int tries = 0;
		
		oreWhileLoop:
		while(nodes.size() < amount && tries < 10) {
			Int3D node = nodes.get(random.nextInt(nodes.size()));
			
			neighbors.clear();
			neighbors.add(new Int3D(node.x+1, node.y, node.z));
			neighbors.add(new Int3D(node.x-1, node.y, node.z));
			neighbors.add(new Int3D(node.x, node.y+1, node.z));
			neighbors.add(new Int3D(node.x, node.y-1, node.z));
			neighbors.add(new Int3D(node.x, node.y, node.z+1));
			neighbors.add(new Int3D(node.x, node.y, node.z-1));
			Collections.shuffle(neighbors);
			
			for(Int3D n : neighbors) {
				if(!nodes.contains(n)) {
					if(getBlockSafe(n.x, n.y, n.z) == stone) {
						nodes.add(n);
						tries = 0;
						continue oreWhileLoop;
					}
				}
			}
			tries++;
		}
		
		for(Int3D n : nodes) {
			setBlock(n.x, n.y, n.z, material);
		}
		
	}
	
}
