package aerogen;

import java.util.ArrayList;
import java.util.Random;

import org.bukkit.World;

public class SchematicSpawnHandler {

	private ArrayList<Schematic> schematics = new ArrayList<Schematic>();
	private ArrayList<Schematic> remove = new ArrayList<Schematic>();
	
	private Random random = new Random();
	private int chunkMinX, chunkMaxX;
	private int chunkMinZ, chunkMaxZ;
	private boolean[][] chunkLoaded = null;
	private boolean[][] chunkPrepared = null;
	
	private AeroGenerator gen;
	
	public SchematicSpawnHandler(AeroGenerator gen) {
		this.gen = gen;
	}
	
	public byte[] generate(World world, int chunkX, int chunkZ) {
		loadChunk(world, chunkX, chunkZ);
		
		byte[] blocks = new byte[32768];
		
		remove.clear();
		for(Schematic scheme : schematics) {
			boolean schemeComplete = scheme.addToChunk(blocks, chunkX, chunkZ);
			if(schemeComplete) remove.add(scheme);
		}
		for(Schematic scheme : remove) {
			removeSchematic(scheme);
		}
				
		return blocks;		
	}
	
	private void loadChunk(World world, int chunkX, int chunkZ) {
		for(int i = chunkX-Schematic.maxSchematicChunkSize; i <= chunkX+Schematic.maxSchematicChunkSize; i++) {
			for(int k = chunkZ-Schematic.maxSchematicChunkSize; k <= chunkZ+Schematic.maxSchematicChunkSize; k++) {
				prepareChunk(world, i, k);
			}
		}
		chunkLoaded[chunkX-chunkMinX][chunkZ-chunkMinZ] = true;
	}

	private void addSchematic(Schematic scheme) {
		schematics.add(scheme);
		AeroGen.log("Schematic created: "+schematics.size());
	}
	
	private void removeSchematic(Schematic scheme) {
		schematics.remove(scheme);
		AeroGen.log("Schematic finished: "+schematics.size());
	}

	public boolean isChunkLoaded(int chunkX, int chunkZ) {
		if(chunkLoaded == null) return false;
		if(chunkX < chunkMinX || chunkX > chunkMaxX || chunkZ < chunkMinZ || chunkZ > chunkMaxZ) return false;
		return chunkLoaded[chunkX][chunkZ];
	}
	
	public void prepareChunk(World world, int chunkX, int chunkZ) {
		if(chunkPrepared == null) createChunkArray(chunkX, chunkZ);
		while(chunkX < chunkMinX || chunkX > chunkMaxX || chunkZ < chunkMinZ || chunkZ > chunkMaxZ) {
			expandChunkArray(chunkX, chunkZ);
		}
		
		if(chunkPrepared[chunkX-chunkMinX][chunkZ-chunkMinZ]) return;
		
		long seed = (long) (chunkX*world.getSeed()) % (Long.MAX_VALUE/2);
		seed += (long) (chunkZ*(world.getSeed() % 65536)*512);
		random.setSeed(seed);
		
		Island island = Island.createIsland(gen, world, random, chunkX, chunkZ);
		if(island != null) {
			checkCompletedChunks(island);
			addSchematic(island);
		}
		
		chunkPrepared[chunkX-chunkMinX][chunkZ-chunkMinZ] = true;
	}

	private void checkCompletedChunks(Schematic scheme) {
		for(int i = chunkMinX; i <= chunkMaxX; i++) {
			for(int k = chunkMinZ; k <= chunkMaxZ; k++) {
				if(chunkLoaded[i-chunkMinX][k-chunkMinZ]) scheme.completeChunk(i, k);
			}
		}
	}

	private void createChunkArray(int chunkX, int chunkZ) {
		chunkMinX = chunkX;
		chunkMinZ = chunkZ;
		chunkMaxX = chunkX;
		chunkMaxZ = chunkZ;
		chunkLoaded = new boolean[1][1];
		chunkPrepared = new boolean[1][1];
	}

	private void expandChunkArray(int x, int y) {
		int xOffset = 0;
		int	zOffset = 0;
		int xSize = chunkMaxX - chunkMinX + 1;
		int	zSize = chunkMaxZ - chunkMinZ + 1;
		int newMinX = chunkMinX, newMaxX = chunkMaxX;
		int	newMinZ = chunkMinZ, newMaxZ = chunkMaxZ;
		
		if(x < chunkMinX) { xOffset = xSize;	newMinX -= xSize; }
		if(y < chunkMinZ) { zOffset = zSize; 	newMinZ -= zSize; }
		if(x > chunkMaxX) 	newMaxX += xSize; 
		if(y > chunkMaxZ) 	newMaxZ += zSize; 
		
		//Copy all the array data
		boolean[][] newChunkArray = new boolean[newMaxX - newMinX + 1][newMaxZ - newMinZ + 1];
		for(int copyX = chunkMinX; copyX <= chunkMaxX; copyX++) {
			System.arraycopy(chunkLoaded[copyX - chunkMinX], 0, newChunkArray[copyX - chunkMinX + xOffset], zOffset, zSize);
		}
		
		//Copy all the array data
		boolean[][] newChunkArrayP = new boolean[newMaxX - newMinX + 1][newMaxZ - newMinZ + 1];
		for(int copyX = chunkMinX; copyX <= chunkMaxX; copyX++) {
			System.arraycopy(chunkPrepared[copyX - chunkMinX], 0, newChunkArrayP[copyX - chunkMinX + xOffset], zOffset, zSize);
		}
		
		chunkLoaded = newChunkArray;
		chunkPrepared = newChunkArrayP;
		chunkMinX = newMinX;
		chunkMinZ = newMinZ;
		chunkMaxX = newMaxX;
		chunkMaxZ = newMaxZ;
	}
	
}
